import torch.nn as nn


def calc_module_params(module: nn.Module) -> int:
    """
    Get the number of parameters in a module
    """
    tot_params = 0
    for param in module.parameters():
        tot_params += param.nelement()
    return tot_params



from typing import Tuple


class MetricBase(object):
    def __init__(self) -> None:
        # prediction history for calculating metric scores
        self.preds = []
        # gold history for calculating metric scores
        self.golds = []

    def reset(self):
        """
        Clear history, often used in the end of one epoch's evaluation
        """
        self.preds.clear()
        self.golds.clear()

    def get_results(self):
        """Get evaluation results from history"""
        return self.calculate_scores(self.golds, self.preds)

    def __call__(self, raw_batch: dict, out_batch: dict) -> dict:
        """
        Calculate evaluation metric scores of current batch
            and add pred & gold results to history.

        Args:
            raw_batch: raw input batch from data loader
            out_batch: predicted results of current batch from model

        Returns:
            a dict containing neccessary information from
        """
        gold_instances, pred_instances = self.get_instances_from_batch(
            raw_batch, out_batch
        )

        self.golds.extend(gold_instances)
        self.preds.extend(pred_instances)

        batch_result = {
            "gold": gold_instances,
            "pred": pred_instances,
            "metric_scores": self.calculate_scores(gold_instances, pred_instances),
        }

        return batch_result

    def get_instances_from_batch(self, raw_batch: dict, out_batch: dict) -> Tuple:
        """
        Process raw batch input and predicted output into
            gold & pred instances

        Args:
            raw_batch: raw input batch from data loader
            out_batch: predicted results of current batch from model

        Returns:
            gold_instances: list of instances
            pred_instances: list of pred instances
        """
        raise NotImplementedError

    def calculate_scores(self, golds: list, preds: list) -> dict:
        """
        Calculate scores from gold and pred instances

        Args:
            golds: a list of gold instances obtained from `get_instances_from_batch()`
            preds: a list of predicted instances

        Returns:
            a dict of metric scores
        """
        raise NotImplementedError



from typing import Tuple

import torch

from rex.metrics.base import MetricBase
from rex.tasks.simple_task import SimpleTask
from rex.utils.dict import get_dict_content
from rex.utils.io import dump_json, dump_jsonlines
from rex.utils.logging import logger
from rex.utils.progress_bar import pbar
from rex.utils.registry import register


@register("task")
class SimpleMetricTask(SimpleTask):
    def __init__(self, config, **kwargs) -> None:
        super().__init__(config, **kwargs)

    def after_initialized(self):
        super().after_initialized()

        logger.debug("Init metric")
        self.metric = self.init_metric()

    def init_metric(self) -> MetricBase:
        raise NotImplementedError

    @torch.no_grad()
    def eval(
        self, dataset_name, verbose=False, dump=False, dump_middle=True, postfix=""
    ) -> Tuple[float, dict]:
        """Eval on specific dataset and return loss and measurements

        Args:
            dataset_name: which dataset to evaluate
            verbose: whether to log evaluation results
            dump: if True, dump metric results to `self.measures_path`
            dump_middle: if True, dump middle results to `self.middle_path`
            postfix: filepath postfix for dumping

        Returns:
            eval_loss: float
            metrics: dict
        """
        self.model.eval()
        eval_loader = self.get_data_loader(
            dataset_name, is_eval=True, epoch=self.history["curr_epoch"]
        )
        loader = pbar(eval_loader, desc=f"{dataset_name} - {postfix} Eval", ascii=True)

        eval_loss = 0.0
        tot_batch_results = []
        for batch in loader:
            out = self.model(**batch)
            eval_loss += out["loss"].item()
            batch_results: dict = self.metric(batch, out)
            batch_metric_score = get_dict_content(
                batch_results["metric_scores"], self.config.best_metric_field
            )
            loader.set_postfix({self.config.best_metric_field: batch_metric_score})

            tot_batch_results.append(batch_results)

        logger.info(loader)
        measurements = self.metric.get_results()

        if verbose:
            logger.info(f"Eval dataset: {dataset_name}")
            logger.info(f"Eval loss: {eval_loss}")
            logger.info(
                f"Eval metrics: {get_dict_content(measurements, self.config.best_metric_field)}"
            )
        _filename_prefix = (
            f"{dataset_name}.{postfix}" if len(postfix) > 0 else f"{dataset_name}"
        )
        if dump:
            dump_obj = {
                "dataset_name": dataset_name,
                "eval_loss": eval_loss,
                "metrics": measurements,
            }
            _measure_result_filepath = self.measures_path.joinpath(
                f"{_filename_prefix}.json"
            )
            dump_json(dump_obj, _measure_result_filepath)
            logger.info(f"Dump measure results into {_measure_result_filepath}")
        if dump_middle:
            _middle_result_filepath = self.middle_path.joinpath(
                f"{_filename_prefix}.jsonl"
            )
            dump_jsonlines(tot_batch_results, _middle_result_filepath)
            logger.info(f"Dump middle results into {_middle_result_filepath}")

        self.metric.reset()

        return eval_loss, measurements



from typing import Dict, Iterable, List, Optional


def validate_instance_has_the_same_keys(data: List[dict]) -> set:
    """Validate if all data have the same keys"""
    all_keys = set()
    for d in data:
        if len(all_keys) == 0:
            all_keys.update(d.keys())
        else:
            assert (
                set(d.keys()) == all_keys
            ), "Data instances does not have the same keys!"
    return all_keys


def group_instances_into_batch(
    instances: List[dict], keys: Optional[Iterable] = None
) -> dict:
    if not keys:
        keys = validate_instance_has_the_same_keys(instances)

    final_data = {key: [] for key in keys}
    for ins in instances:
        for key in keys:
            final_data[key].append(ins[key])
    return final_data


def decompose_batch_into_instances(batch: Dict[str, list]) -> List[dict]:
    keys = batch.keys()
    batch_size = 0
    for key in keys:
        batch_size = len(batch[key])
        break

    instances = []
    for i in range(batch_size):
        ins = {key: batch[key][i] for key in keys}
        instances.append(ins)

    return instances
